{{ sdram-40MHz.spin }}
CON
  Q_SIZE = 400
  Q_BYTES = Q_SIZE << 3   ' entries are 2 longs

  WAITCYCLES = 36
  WIDTH = 800
  HEIGHT = 480
  PAGE_HEIGHT = 512
  HGAP = 100      ' pixels per hblank and vblank
  VGAP = 100


  pin_BA0  = 13
  pin_BA1  = 14

  pin_LE   = 16
  pin_DQM  = 17  ' waitvid controlled
  pin_WE   = 18
  pin_CAS  = 19
  pin_RAS  = 20
  pin_CS   = 21
  pin_CKE  = 22  ' waitvid controlled
  pin_DEN  = 23

  pin_SCLK = 24
  pin_DCLK = 25
  pin_LCLK = 26

  DENPIN  = %10000000   ' these are video byte masks
  CEPIN   = %01000000
  CSPIN   = %00100000
  RASPIN  = %00010000
  CASPIN  = %00001000
  WEPIN   = %00000100
  DQMPIN  = %00000010
  LEPIN	  = %00000001


  IDLE       =	CSPIN | RASPIN | CASPIN | WEPIN | LEPIN
  NONE	     =	CEPIN | RASPIN | CASPIN | WEPIN | LEPIN
  DEN        =	DENPIN | CEPIN | RASPIN | CASPIN | WEPIN | LEPIN
  LE         =	CEPIN | RASPIN | CASPIN | WEPIN 
  LE_DQM     =	CEPIN | RASPIN | CASPIN | WEPIN | DQMPIN
  DQM        =	CEPIN | RASPIN | CASPIN | WEPIN | LEPIN | DQMPIN
  ACTIVE     =	CEPIN | CASPIN | WEPIN | LEPIN
  ACTIVE_DEN =	DENPIN | CEPIN | CASPIN | WEPIN | LEPIN
  READ	     =	CEPIN | RASPIN | WEPIN | LEPIN
  READ_DEN   =	DENPIN | CEPIN | RASPIN | WEPIN | LEPIN
  WRITE	     =	CEPIN | RASPIN | LEPIN
  WRITE_LE   =	CEPIN | RASPIN
  WRITE_LE_DQM=	CEPIN | RASPIN | DQMPIN
  BURST_TERM =	CEPIN | RASPIN | CASPIN | LEPIN
  BURST_TERM_DEN = DENPIN | CEPIN | RASPIN | CASPIN | LEPIN
  PRECHARGE  =	CEPIN | CASPIN | LEPIN
  PRECHARGE_DEN= DENPIN | CEPIN | CASPIN | LEPIN
  AUTO_REFR  =	CEPIN | WEPIN | LEPIN
  LOADREG    =	CEPIN | LEPIN


  OP_NOP   = 0
  OP_HSPAN = 1
  OP_VSPAN = 2
  OP_CHAR  = 3
  OP_SPRITE = 4
  OP_INVCHAR = 5
  OP_CLIPRECT = 6

OBJ
  poly:   "sdram-video-raster"
  circle: "sdram-video-circles"


VAR
'  long lock
  long queue [Q_SIZE*2]
'  word edge_table [MAX_EDGES * EDGE_SIZE]  ' about 2k RAM

  long sdram_cog
'  long polygon_cog

  ' arguments to PASM, must stay in this sequence
  long arg_queue        ' passed to all cogs
  long arg_offsets
  long qp		' passed by address to all cogs
  long arg_vcount       ' offset 3, managed by raster
  long arg_poly_colour  ' managed by raster
  long arg_vtable       ' managed by raster
  long arg_etable       ' offset 6, managed by raster
  long arg_circ_radcol  ' offset 7, managed by circle
  long arg_circ_xy      ' managed by circle

  long write_voffset
  long display_page

  long foreground

  word voffset
  word hoffset


PUB  Start
  Stop

  display_page := 0
  write_voffset := 0
  arg_queue := @queue
  arg_offsets := @voffset      ' leftmost col_addr
  init_queues

{ Launch SDRAM cog }
  sdram_cog   := 1 + cognew (@entry,   @arg_queue) 'Launch new cog
  if sdram_cog == 0
    return false
  if !poly.Start (@arg_queue)
    Stop
    return false
  if !circle.Start (@arg_queue)
    poly.Stop
    Stop
    return false
  return true

PUB  Stop
  poly.Stop
  circle.Stop
  if sdram_cog <> 0
    cogstop (sdram_cog-1)
    sdram_cog := 0 

{ queue handling code }
PUB init_queues | i
  qp := 0
  repeat i from 0 to Q_SIZE-1
    queue [i*2] := 0


PUB  drain_queues | i
  repeat until arg_circ_radcol == 0
  repeat until arg_vcount == 0
  repeat i from 0 to Q_SIZE-1
    repeat until queue [i*2] == 0

PUB  set_display_page (page)
  display_page := page
  set_display_offsets (0, 0)

PUB  set_display_offsets (left, top)
  voffset := PAGE_HEIGHT * display_page + top
  hoffset := left

PUB  set_write_page (page)
  write_voffset := PAGE_HEIGHT * page
  set_cliprect (0, 0, WIDTH, HEIGHT)

PUB  set_page (page)
  set_write_page (page)
  set_display_page (page)

PUB  set_colour (red, green, blue)
  foreground := (blue & $F8) << 8 | (green & $FC) << 3 | (red & $F8) >> 3
 
PUB  set_cliprect (left, top, right, bot)
  enqueue (OP_CLIPRECT, left, top, right, bot+write_voffset)

PUB  enqueue (tag, x, y, param, col)
  repeat until queue [qp] == 0   ' await free slot (should lock)
  queue [qp+1] := (param << 16) | (col & $FFFF)
  queue [qp] := (tag << 29) | ((x & $1FFF) << 16) | ((y+write_voffset) & $FFFF)
  qp += 2
  if qp => 2*Q_SIZE
    qp := 0

{
PUB  enqueue (tag, x, y, param, col)
  repeat
    repeat until not lockset (lock)
    if queue [qp] == 0
      queue [qp+1] := (param << 16) | (col & $FFFF)
      queue [qp] := (tag << 29) | ((x & $1FFF) << 16) | (y & $7FFF)
      qp += 2
      if qp => 2*Q_SIZE
        qp := 0
      lockclr (lock)
      return
    lockclr (lock)
}

PUB  draw_poly (vertex_count, vertex_table) | i
  if vertex_count > 2
'    repeat until arg_vcount == 0 ' wait to finish
    repeat i from 0 to vertex_count-1
      long [vertex_table][i] += write_voffset<<16
    arg_vtable := vertex_table
    arg_poly_colour := foreground
    arg_vcount := vertex_count | $80000000  ' triggers

    repeat until arg_vcount == 0 ' wait to finish
    repeat i from 0 to vertex_count-1
      long [vertex_table][i] -= write_voffset<<16

PUB  fill_poly (vertex_count, vertex_table) | i  ' how to pass write offset?
  if vertex_count > 2
'    repeat until arg_vcount == 0 ' wait to finish
    repeat i from 0 to vertex_count-1
      long [vertex_table][i] += write_voffset<<16
    arg_vtable := vertex_table
    arg_poly_colour := foreground
    arg_vcount := vertex_count   ' triggers

    repeat until arg_vcount == 0 ' wait to finish
    repeat i from 0 to vertex_count-1
      long [vertex_table][i] -= write_voffset<<16


PUB  draw_span (x, y, wid)
  enqueue (OP_HSPAN, x, y, wid, foreground)

PUB  draw_vline (x, y, h)
  enqueue (OP_VSPAN, x, y, h, foreground)

PUB  draw_dot (x, y)
  enqueue (OP_VSPAN, x, y, 1, foreground)


PUB  draw_glyph (x, y, chr)
  enqueue (OP_CHAR, x, y, chr, foreground)
PUB  draw_invglyph (x, y, chr)
  enqueue (OP_INVCHAR, x, y, chr, foreground)

PUB  draw_sprite (x, y, addr)
  enqueue (OP_SPRITE, x, y, addr, foreground)

{{
PUB  fill_rect (left, top, right, bot) | i, j, w
  if left < right and top < bot
    if right-left > bot-top
      w := right-left+1
      repeat i from top to bot
        draw_span (left, i, w)
    else
      w := bot-top+1
      repeat until w =< 64
        repeat i from left to right
          draw_vline (i, top, 64)
        w -= 64
        top += 64
      repeat i from left to right
        draw_vline (i, top, w)
}}

PUB  fill_rect (left, top, right, bot) | i, j, w
  if left < right and top < bot
    w := right-left+1
    repeat i from top to bot
      draw_span (left, i, w)


PUB  draw_rect (left, top, right, bot) | i, j, w, h
  if left < right and top < bot
    w := right-left+1
    draw_span (left, top, w)
    draw_span (left, bot, w)
    h := bot-top-1
    if h == 0
      return
    repeat until h =< 64
      draw_vline (left, top+1, 64)
      draw_vline (right, top+1, 64)
      top += 64
      h -= 64
    draw_vline (left, top+1, h)
    draw_vline (right, top+1, h)

   

PUB  border_rect (left, top, right, bot) | pcol
  pcol := foreground
  fill_rect (left, top, right, bot)
  set_colour (0,0,0)
  draw_rect (left, top, right, bot)
  foreground := pcol


PUB  draw_line (x0, y0, x1, y1) | dx, xstep, dy, ystep, x, y, t
  if x1 > x0
    dx := x1 - x0
    xstep := 1
  else
    dx := x0 - x1
    xstep := -1
  if y1 > y0
    dy := y1 - y0
    ystep := 1
  else 
    dy := y0 - y1
    ystep := -1
  x := x0
  y := y0
  if dx < dy
    t := -(dy>>1)
    repeat
      draw_dot (x, y)
      if y == y1
        return
      y += ystep
      t += dx
      if t => 0
        x += xstep
        t -= dy
  else
    t := -(dx>>1)
    repeat
      draw_dot (x, y)
      if x == x1
        return
      x += xstep
      t += dy
      if t => 0
        y += ystep
        t -= dx

PUB  border_circle (x, y, radius) | pcol
  pcol := foreground
  fill_circle (x, y, radius)
  set_colour (0, 0, 0)
  draw_circle (x, y, radius)
  foreground := pcol

PUB  fill_circle (x, y, radius) 
  if radius > 0
    repeat until arg_circ_radcol == 0
    arg_circ_xy := (x & $FFFF) << 16 | ((y+write_voffset) & $FFFF)
    arg_circ_radcol := $80000000 | radius << 16 | (foreground & $FFFF)
    repeat until arg_circ_radcol == 0

PUB  draw_circle (x, y, radius) 
  if radius > 0
    repeat until arg_circ_radcol == 0
    arg_circ_xy := (x & $FFFF) << 16 | ((y+write_voffset) & $FFFF)
    arg_circ_radcol := (radius & $7FFF) << 16 | (foreground & $FFFF)
    repeat until arg_circ_radcol == 0

{{
PUB  slow_fill_circle (x, y, radius) | f, delx, dely, x2, y2  ' circle drawing code
  if radius < 1
    return
  f := -radius
  delx := 1
  dely := radius+radius+1
  x2 := 0
  y2 := radius
  repeat
    repeat while f < -delx
      x2 ++
      f += delx
      delx += 2
    circle_slice (x-x2, x+x2, y-y2)  ' only bother if y will change next step
    circle_slice (x-x2, x+x2, y+y2)
    dely -= 2
    f -= dely
    y2 --
  until y2 == 0
  circle_slice (x-radius, x+radius, y)
}}

PUB  fill_ellipse (x, y, xrad, yrad) | f, delx, dely, x2, y2, alp, bet  ' circle drawing code
  if xrad < 1 or yrad < 1
    return
  alp := xrad*xrad
  bet := yrad*yrad
  f := -(xrad*yrad*(xrad+yrad) >> 1)
  delx := bet
  dely := (yrad+yrad+1) * alp
  x2 := 0
  y2 := yrad
  repeat
    repeat while f < -delx
      x2 ++
      f += delx
      delx += 2*bet
    circle_slice (x-x2, x+x2, y-y2)  ' only bother if y will change next step
    circle_slice (x-x2, x+x2, y+y2)
    dely -= 2*alp
    f -= dely
    y2 --
  until y2 == 0
  circle_slice (x-xrad, x+xrad, y)


PRI  circle_slice (xl, xh, y)
  draw_span (xl, y, xh-xl+1)


{{
PUB  slow_draw_circle (x, y, radius) | f, delx, dely, x2, y2, xp   ' circle drawing code
  if radius < 1
    return
  f := radius
  delx := 1
  dely := radius+radius+1
  x2 := 0
  xp := 0
  y2 := radius
  draw_dot (x-radius, y)
  draw_dot (x+radius, y)
  draw_dot (x, y-radius)
  draw_dot (x, y+radius)
'  draw_dot (x-radius, y)  ' hack for slow colour setup
  repeat
    repeat while f => delx
      x2 ++
      f -= delx
      delx += 2
    circle_chunk (x, y, x2, y2, x2-xp)
    xp := x2
    dely -= 2
    f += dely
    y2 --
  until y2 < x2


PRI  circle_chunk (x, y, x2, y2, w)
  if w > 0
    w--
    draw_span  (x-x2,   y-y2, w+1)
    draw_span  (x-x2,   y+y2, w+1)
    draw_span  (x+x2-w, y-y2, w+1)
    draw_span  (x+x2-w, y+y2, w+1)
    draw_vline (x-y2,   y-x2, w+1)
    draw_vline (x+y2,   y-x2, w+1)
    draw_vline (x-y2, y+x2-w, w+1)
    draw_vline (x+y2, y+x2-w, w+1)
}}


DAT
		ORG	0

entry           mov     parm, PAR
		rdlong  queue_beg, parm
		add	parm, #4
		rdlong	scanbase, parm		' parm left pointing at scanbase

		mov	queuep, queue_beg
		mov	queue_end, queue_beg
		add	queue_end, queue_bytes


                mov     OUTA, #0
                or	DIRA, LOW_WORD
                or      DIRA, CKEMASK           ' ensure CKE driven LOW
		or	DIRA, LCLKMASK          ' and LCLK

                mov	FRQA, CLOCK_FREQ
                mov     PHSA, PHASE
		movs	CTRA, #pin_SCLK
		movi	CTRA, #%0_00010_101	' PLL / 4, so 20MHz

                mov	FRQB, CLOCK_FREQ
                mov     PHSB, PHASE
		movs	CTRB, #pin_DCLK
		movi	CTRB, #%0_00010_101     ' change 101 to 100 for halfspeed LCD clock

		mov	VSCL, V4096
		mov	VCFG, vconfig_go
		WAITVID	idle_comms, #0	' force the video hardware to a known state
                or      DIRA, VIDMASK
                or      DIRA, SCLK_MASK		' now clocking in inactive state
                or      DIRA, DCLK_MASK

		mov	temp, CNT
		add	temp, hundred_us
		waitcnt	temp, delay
		call 	#init_sdram

top_of_frame
:loop           mov     count, Hheight
                mov     VSCL, V25
                rdlong  row_addr, scanbase
		mov	col_addr, row_addr

 		WAITVID init_comms, init_seq   ' paranoid - reset mode word
		mov	OUTA, precharge_all
                nop
                mov	OUTA, modeword

		shr	col_addr, #16
		and	col_addr, COLADDR_MASK
	        and     row_addr, ADDR_MASK

                mov     bank_addr, row_addr
                and     bank_addr, BANK_MASK

                or      col_addr, bank_addr  ' was mov
:scanloop
                WAITVID refresh_comms, refresh_seq  ' synchronize with end of blanking
                call    #burst_read
                add     row_addr, #1
                and     row_addr, ADDR_MASK
                WAITVID refresh_comms, refresh_seq
                mov     bank_addr, row_addr
                and     bank_addr, BANK_MASK
                WAITVID refresh_comms, refresh_seq
		and	col_addr, COLADDR_MASK
		or	col_addr, bank_addr'  mov     col_addr, bank_addr   ' assume in one bank
                djnz    count, #:scanloop

                WAITVID refresh_comms, refresh_seq
                mov     cycles_left, GAP_CYCLES
                mov     VSCL, Vwait
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
                jmp     #span_interp



refresh_comms   long    NONE | AUTO_REFR<<8
refresh_seq     long    %00_01_00_00_00_00_00_01_00_00

ba0_mask	long	1 << pin_BA0
ba1_mask	long	1 << pin_BA1
BANK_MASK       long    3 << pin_BA0
ADDR_MASK       long    $00007FFF

VIDMASK		long	$00FF0000
CKEMASK         long    1 << pin_CKE
LCLKMASK	long	1 << pin_LCLK
SCLK_MASK	long	1 << pin_SCLK
DCLK_MASK       long    1 << pin_DCLK
vconfig_stop	long    %0_00_1_0_0_000_00000_000000010_011111111  ' 2 bits/pixel VGA mode
vconfig_go	long    %0_01_1_0_0_000_00000_000000010_011111111


' single burst
burst_read      mov     VSCL, V800
                WAITVID read_comms1, read_seq1
                mov     OUTA, row_addr   ' ACTIVE
                mov     OUTA, col_addr    ' READ
                mov     VSCL, V25
                WAITVID read_comms2, read_seq2
		mov     OUTA, bank_addr     ' PRECHARGE bank
burst_read_ret  ret

{{
' switch-banks burst, allows more pixels per line
burst_read      mov     VSCL, V200
                WAITVID read_comms1, read_seq1
                mov     OUTA, row_addr   ' ACTIVE
                mov     OUTA, col_addr    ' READ
		add	row_addr, ba0_mask
		add	col_addr, ba0_mask
		
		mov	bcount, #3
:burst_loop
		WAITVID	read_commsx, read_seqx
		mov	OUTA, row_addr
		mov	OUTA, col_addr
		nop
		mov	OUTA, bank_addr
		add	row_addr, ba0_mask
		add	col_addr, ba0_mask
		add	bank_addr, ba0_mask
		djnz	bcount, #:burst_loop

                mov     VSCL, V25
                WAITVID read_comms2, read_seq2
		mov     OUTA, bank_addr     ' PRECHARGE bank
		sub	bank_addr, BANK_MASK
		and	row_addr, ADDR_MASK
		and	col_addr, ADDR_MASK
burst_read_ret  ret

bcount		long	0

V200            long    $1000 | 200
V300            long    $1000 | 300
V500            long    $1000 | 500
}}

{
' multi burst, allows horizontal scroll (upto a point)
burst_read      mov     VSCL, Vhscroll1
		add	col_addr, hscroll
                WAITVID read_comms1, read_seq1
                mov     OUTA, row_addr   ' ACTIVE
                mov     OUTA, col_addr    ' READ
		sub	col_addr, hscroll

		mov	VSCL, Vhscroll2
		WAITVID	read_commsx, read_seqxx
		nop
		mov	OUTA, col_addr

                mov     VSCL, V25
                WAITVID read_comms2, read_seq2
		mov     OUTA, bank_addr     ' PRECHARGE bank
burst_read_ret  ret

hscroll		long    scroll_amount
Vhscroll1	long	$1000 | (WIDTH - scroll_amount)
Vhscroll2	long	$1000 | scroll_amount
}


read_comms1     long    NONE | ACTIVE<<8 | READ<<16 | DEN<<24
read_seq1       long    %11_11_11_11_11_11_11_11_11_11_00_10_00_01_00_00

read_commsx	long	DEN | ACTIVE_DEN<<8 | READ_DEN<<16 | PRECHARGE_DEN<<24
read_seqx	long	%00_11_00_00_00_10_00_01_00_00
read_seqxx	long	%00_10_00_00_00_00


read_comms2     long    NONE | BURST_TERM_DEN<<8 | PRECHARGE<<16 | DEN<<24
read_seq2       long    %00_10_00_00_11_01_11_11_11_11





write_comms0    long    NONE | ACTIVE<<8 | WRITE_LE<<16 | BURST_TERM<<24
write_seq00     long    %11_10_00_01_00_00
write_seq0      long    %01_00_00 ' nop, active

write_comms1    long    NONE | LE<<8 | BURST_TERM<<16 | WRITE_LE<<24
write_seq1      long    %01_01_01_01_01_01_01_01_01_01_01_01_01_01_01_01  ' all LE
new_nar_seq	long	%10_01_01_01_01_01_01_01_01_01_01_01_01_01_01_01
new_wide_seq	long	%01_01_01_01_01_01_01_01_01_01_01_01_01_01_01_11


write_comms2    long    NONE | BURST_TERM<<8 | PRECHARGE<<16  ' $4A5A5E  ' NONE, BT, PRECHARGE
write_seq2      long    %00_10_00_01


sprite_setup    WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
         
                rdlong  patt, char		' first charline outside loop
		add	char, #4

                movs    VSCL, #32
                WAITVID char_comms0, char_seq0   ' setup ACTIVE
                mov     OUTA, row_addr

		mov	temp, patt
                and     patt, H55555555		' even bits first
	  	or	patt, #2
		shr	temp, #1
		and	temp, H55555555		' odd bits

		mov	OUTA, colour
                mov     count, #32		' 32x32 sprites  - could the count be an arg?
		movs	VSCL, #16
                or      OUTA, LCLKMASK

:cloop
                mov     OUTA, col_addr
                WAITVID char_comms1, patt
		rdlong	patt, char		' takes 23 cycles worst case
		WAITVID char_comms1, temp
		add	char, #4
		mov	temp, patt
                and     patt, H55555555		' even bits first
	  	or	patt, #2
		shr	temp, #1
		and	temp, H55555555		' odd bits
                WAITVID char_comms2, char_seq2  ' BT/PRECHARGE/NOP/ACTIVE
                add     row_addr, #1            ' ready for next row
                mov     OUTA, bank_addr            ' precharge
                mov     OUTA, row_addr          ' active
                djnz    count, #:cloop

                mov     VSCL, Vwait
                WAITVID char_comms2, char_seq2x
                mov     OUTA, bank_addr     ' precharge
                sub     cycles_left, sprite_cycles
sprite_setup_ret ret

sprite_cycles	long	32 + 32 * (16+16+16) + WAITCYCLES


char_setup      WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
                and     char, #$FF
         
		mov	OUTA, colour

		ror	char, #1  wc		' Prop font address decoding, C flag has odd/even
                shl     char, #7
                add     char, fontbase
                or      OUTA, LCLKMASK

                movs    VSCL, #32
                WAITVID char_comms0, char_seq0   ' setup ACTIVE
                mov     OUTA, row_addr

                rdlong  patt, char		' first charline outside loop
		add	char, #4
	if_c    shr     patt, #1
                and     patt, H55555555
	  	or	patt, #2

                mov     count, #32		' charline count
		movs	VSCL, #16
:cloop					' could we skip blank lines?  pass address and line count, not char code?
                mov     OUTA, col_addr
                WAITVID char_comms_var, patt

                rdlong  patt, char

                WAITVID char_comms2, char_seq2  ' BT/PRECHARGE/NOP/ACTIVE
                add     row_addr, #1            ' ready for next row
                mov     OUTA, bank_addr         ' precharge
                mov     OUTA, row_addr          ' active
		add	char, #4

	if_c    shr     patt, #1		' select odd or even character
		WAITVID nop_comms, #0
                and     patt, H55555555	        ' mask out other
	  	or	patt, #2                ' insert WRITE bit

                djnz    count, #:cloop

                mov     VSCL, Vwait
                WAITVID char_comms2, char_seq2x
                mov     OUTA, bank_addr     ' precharge
                sub     cycles_left, char_cycles
char_setup_ret  ret

char_cycles	long	32 + 32 * (16+16+16) + WAITCYCLES


char            long    0
patt            long    0
fontbase        long    $8000
H55555555       long    $55555555
D2              long    2 << 9
V9              long    $1000 | 9
V16             long    $1000 | 16
V24 		long	$1000 | 24

char_comms0     long    NONE | ACTIVE<<8  ' $4E5E      ' NONE, ACTIVE
char_seq0       long    %00_00_01_00_00

char_comms1     long    LE_DQM | LE<<8 | WRITE_LE_DQM<<16 | WRITE_LE<<24
char_comms1inv  long    LE | LE_DQM<<8 | WRITE_LE<<16 | WRITE_LE_DQM<<24
char_comms_var	long	0

char_comms2     long    NONE | BURST_TERM<<8 | PRECHARGE<<16 | ACTIVE<<24
char_seq2       long    %00_11_00_00_00_10_00_00_00_01
char_seq2x      long    %00_10_00_00_00_00


idle_comms  	long	IDLE
nop_comms       long    NONE

modeword	long	%00_0_00_010_0_111
precharge_all	long	$400		' A10 set

init_comms	long	NONE | PRECHARGE<<8 | LOADREG<<16    ' $424A5E ' NONE ,PRECHARGE,LOADREG
init_seq	long	%00_10_00_00_00_01_00_00


init_sdram     	mov	VSCL, V12
 		WAITVID init_comms, init_seq
		mov	OUTA, precharge_all
                nop
                mov	OUTA, modeword
                mov     VSCL, V25
		WAITVID nop_comms, #0

                mov     count, #1
		shl	count, #13
:loop
                WAITVID refresh_comms, refresh_seq  ' refresh every row
                djnz    count, #:loop

		mov	colour, #0
		mov	OUTA, colour
		nop
		or	OUTA, LCLKMASK
		andn	OUTA, LCLKMASK   ' initialize colour and latch

                WAITVID refresh_comms, refresh_seq  ' avoid waitvid timeout
init_sdram_ret	ret

{{ convert x, y, span, colour args }}

parse_args
		mov	colour, spancol
		and	colour, LOW_WORD     ' LOW_WORD for 16bit colour
		sar	spancol, #16	     ' span might be signed?
		mov	row_addr, xy
		mov	col_addr, xy

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

		shr	col_addr, #16

set_sdram_addresses
		and	row_addr, ADDR_MASK
		mov	bank_addr, row_addr
		and	bank_addr, bank_mask
		and	col_addr, COLADDR_MASK
		or	col_addr, bank_addr
hspan_args_ret
vspan_args_ret
parse_args_ret  ret


hspan_args
		mov	row_addr, xy
		shl	row_addr, #16
		sar	row_addr, #16
		cmps	row_addr, clip_top  wz,wc  ' y clip for hspans
	if_ae	cmps	clip_bot, row_addr  wz,wc

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

	if_be	jmp	#span_interp
		mov	colour, spancol
		and	colour, LOW_WORD     ' LOW_WORD for 16bit colour
		sar	spancol, #16	     ' span might be signed?
		mov	col_addr, xy
		shl	col_addr, #3
		sar	col_addr, #19
		add	spancol, col_addr	' x clip the span
		mins	col_addr, clip_left
		maxs	spancol, clip_right
		cmps	spancol, col_addr  wz,wc,wr

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

	if_a	jmp	#set_sdram_addresses
		jmp	#span_interp

vspan_args
		mov	col_addr, xy
		shl	col_addr, #3
		sar	col_addr, #19
		cmps	col_addr, clip_left  wz,wc
	if_ae	cmps	clip_right, col_addr  wz,wc  ' x clip for vspan

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

	if_be	jmp	#span_interp

		mov	colour, spancol
		and	colour, LOW_WORD     ' LOW_WORD for 16bit colour
		sar	spancol, #16	     ' span might be signed?
		mov	row_addr, xy
		shl	row_addr, #16
		sar	row_addr, #16
		add	spancol, row_addr	' y clip the span
		mins	row_addr, clip_top
		maxs	spancol, clip_bot
		cmps	spancol, row_addr  wz,wc,wr

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

	if_a	jmp	#set_sdram_addresses
		jmp	#span_interp		' this can fall-through...


{{ handle queued graphics operations }}

span_interp	cmp	cycles_left, MIN_CYCLES  wz,wc
	if_b	jmp	#wait_frame          ' run out of time, we're done

		rdlong	xy, queuep  wz	     ' first long of entry is ttt Xxxxxxxxxxxxx Yyyyyyyyyyyyyyyy
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
	if_z	jmp	#span_interp

got_entry
		wrlong	zero, queuep	     ' mark empty
		add	queuep, #4
		mov	temp, xy
		rdlong	spancol, queuep	     ' second long is Ssssssssssssssss cccccccccccccccc 
		and	xy, H1FFFFFFF
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
		add	queuep, #4
		cmp	queuep, queue_end  wz,wc ' wrap queue
	if_ae	mov	queuep, queue_beg

		shr	temp, #29	     ' get tag field
		add	temp, #jump_table
		jmp	temp		     ' indirect

jump_table	jmp	#next_entry
		jmp	#do_hspan            ' OP_HSPAN
		jmp	#do_vspan	     ' OP_VSPAN
		jmp	#do_char	     ' OP_CHAR
		jmp	#do_sprite	     ' OP_SPRITE
		jmp	#do_invchar	     ' OP_INVCHAR
		jmp	#do_clip	     ' OP_CLIPRECT
		jmp	#next_entry

next_entry
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
		jmp	#span_interp

do_clip		' left top right bot
		mov    clip_left, xy
		sar    clip_left, #16
		mov    clip_top, xy
		shl    clip_top, #16
		sar    clip_top, #16
		mov    clip_right, spancol

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

		sar    clip_right, #16
		mov    clip_bot, spancol
		shl    clip_bot, #16
		sar    clip_bot, #16

		mins   clip_right, clip_left    ' sanity check,
		mins   clip_bot, clip_top

		mov    clip_right_16, clip_right ' calculate derived values
		sub    clip_right_16, #16
		mov    clip_right_32, clip_right
		sub    clip_right_32, #32
		mov    clip_bot_32, clip_bot
		sub    clip_bot_32, #32

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

		jmp    #span_interp
		
do_hspan
		call	#hspan_args

		test	spancol, H1FF8  wz
 	if_nz	jmp	#:widespan
:narrowspan
		mov	OUTA, colour
		mov	temp, spancol
		mov	patt, new_nar_seq
		or	OUTA, LCLKMASK
		movs	VSCL, #26
		WAITVID write_comms0, write_seq0  ' #%10_00 ' nop, active
		mov	OUTA, row_addr

		and	temp, #7
		add	temp, temp
		sub	temp, #30
		neg	temp, temp
		shr	patt, temp
		or	patt, #%11

		movs	VSCL, #18
		mov	OUTA, col_addr
		WAITVID write_comms1, patt
		mov	VSCL, Vwait
		mov	OUTA, #0
		WAITVID write_comms2, #%10_00
		sub	cycles_left, #WAITCYCLES+26+18
		jmp	#span_interp

:widespan
		movs	VSCL, #12
                WAITVID refresh_comms, refresh_seq
		mov	OUTA, colour
		mov	temp, spancol
		movs	VSCL, #18
		or	OUTA, LCLKMASK
		WAITVID write_comms0, write_seq0 ' nop, active
		mov	OUTA, row_addr
		max	temp, H400 ' current limit to 1024 spans
		mov	VSCL, temp
		sub	cycles_left, temp
		or	VSCL, H1000
		mov	OUTA, col_addr
		WAITVID write_comms1, new_wide_seq 
		mov	VSCL, Vwait
		mov	OUTA, bank_addr
		WAITVID write_comms2, write_seq2
		sub	cycles_left, #WAITCYCLES+12+18
		jmp	#span_interp


do_vspan
		call	#vspan_args

		movs	VSCL, #24
                WAITVID refresh_comms, refresh_seq
		and	spancol, H3FF
		mov	count, spancol

		mov	OUTA, colour
		shl	spancol, #4
		sub	cycles_left, spancol
		shr	spancol, #2
		sub	cycles_left, spancol

		or	OUTA, LCLKMASK
		mov	spancol, count
		movs	VSCL, #10
:loop
		WAITVID write_comms0, write_seq00
		mov	OUTA, row_addr
		mov	OUTA, col_addr
		add	row_addr, #1
		WAITVID write_comms2, #%10_00_00  ' NOP, PRECHARGE
		mov	OUTA, bank_addr
		djnz	count, #:loop

		WAITVID nop_comms, #0
		mov	VSCL, Vwait
                WAITVID refresh_comms, refresh_seq
		sub	cycles_left, #WAITCYCLES+24+10
		jmp	#span_interp



do_invchar
		mov	char_comms_var, char_comms1inv
		jmp	#char_common
do_char
		mov	char_comms_var, char_comms1
char_common
		call	#parse_args
		WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

		cmps	row_addr, clip_top  wz,wc
	if_ae	cmps	clip_bot_32, row_addr  wz,wc
	if_b	jmp	#span_interp
		xor	col_addr, bank_addr
		cmps	col_addr, clip_left  wz,wc
	if_ae	cmps	clip_right_16, col_addr  wz,wc
	if_b	jmp	#span_interp
		xor	col_addr, bank_addr

		mov	char, spancol
		call	#char_setup
		jmp	#span_interp




do_sprite 	call	#parse_args
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

		cmps	row_addr, clip_top  wz,wc
	if_ae	cmps	clip_bot_32, row_addr  wz,wc
	if_b	jmp	#span_interp
		cmps	col_addr, clip_left  wz,wc
	if_ae	cmps	clip_right_32, col_addr  wz,wc
	if_b	jmp	#span_interp

                mov	char, spancol
		and	char, LOW_WORD
		call	#sprite_setup
		jmp	#span_interp


wait_frame      cmps    cycles_left, #0  wz,wc
        if_be   jmp     #top_of_frame
                cmpsub  cycles_left, D2048  wc
        if_c    mov     VSCL, V2048
        if_nc   mov     VSCL, cycles_left
        if_nc   add     VSCL, WAITCYCLES
        if_nc   or      VSCL, V4096
        if_nc   mov     cycles_left, #0
                WAITVID refresh_comms, refresh_seq
                jmp     #wait_frame






hundred_us	long	8000
delay		long	800
queue_bytes	long	Q_BYTES

CLOCK_FREQ      long    $20000000  ' 5MHz * 16 / 4 = 40MHz PLL
PHASE           long    $00000000

' dataval		long	$000000FF ' 8 bit test pattern, may be 16bit later
writelen        long    250


Hheight         long    HEIGHT
COLADDR_MASK    long    $3FF
H3FF		long	$3FF
H400		long	$400
H7FF		long	$7FF
HFFE0		long	$FFE0
H1FF8		long    $1FF8
H1FFFFFFF	long	$1FFFFFFF
LOW_WORD        long    $0000FFFF
HIGH_WORD       long    $FFFF0000
D2048           long    $800

H1000		long	$1000
H7F8		long	$7F8
'V4              long    $1000 | 4
'V5              long    $1000 | 5
V12             long    $1000 | 12
'V7              long    $1000 | 7
'V11             long    $1000 | 11
V25             long    $1000 | 25
'V50             long    $1000 | 50
Vwait           long    $1000 | WAITCYCLES
V800            long    $1000 | 800
V2048           long    $1000 | 2048
V4096           long    $1000 | 4096

GAP_CYCLES      long    (WIDTH+HGAP)*VGAP
MIN_CYCLES      long    2700  ' 3 scanlines
zero            long    0

clip_top	long	0
clip_bot	long	HEIGHT
clip_bot_32	long	HEIGHT-32
clip_left	long	0
clip_right	long	WIDTH
clip_right_16	long	WIDTH-16
clip_right_32	long	WIDTH-32


parm            res     1

queue_beg	res	1
queue_end	res	1
queuep		res	1

xy		res	1
spancol		res	1

temp            res     1
count           res     1
colour          res     1
cycles_left     res     1

row_addr        res     1
col_addr        res     1
bank_addr       res     1
scanbase	res	1

		FIT	$1F0

{{
////////////////////////////////////////////////////////////////////////////////////////////
//                                TERMS OF USE: MIT License
////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this 
// software and associated documentation files (the "Software"), to deal in the Software 
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////////////////
}}
