{{ sdram-video-circles.spin }}
CON
  WIDTH = 800
  HEIGHT = 480

  Q_SIZE = 400
  Q_BYTES = Q_SIZE << 3   ' entries are 2 longs


  OP_NOP   = 0
  OP_HSPAN = 1
  OP_VSPAN = 2
  OP_CHAR  = 3
  OP_SPRITE = 4


VAR
  long circle_cog


PUB  Start (args)
  Stop
  long [args][7] := 0   ' set workqueue empty
  circle_cog := 1 + cognew (@circle_driver, args) 'Launch new cog
  if circle_cog == 0
    Stop
    return false
  else
    return true

PUB  Stop
  if circle_cog <> 0
    cogstop (circle_cog-1)
    circle_cog := 0


DAT

		ORG	0

circle_driver	mov	parm, PAR
		rdlong  outq_beg, parm
		add	parm, #8
		mov	outqp_addr, parm
		add	parm, #20	' parm now pointing at arg_circ_radcol

		mov	outq_end, outq_beg
		add	outq_end, queue_bytes


workloop	rdlong	rad, parm  wz	' long format is fill (1 bit), radius (15 bits), colour (16 bit)
	if_z	jmp	#workloop       ' busy wait for circle
		mov	spanetc, rad
		and	spanetc, LOW_WORD  ' just the colour
		shr	rad, #16
		test	rad, LOW_UWORD  wz
	if_z	jmp	#mark_done

		rdlong  outqp, outqp_addr	' cache the queue insert pointer now we have work
		shl	outqp, #2
		add	outqp, outq_beg

		add	parm, #4
		rdlong	x0y0, parm
		sub	parm, #4
		mov	y0, x0y0
		shl	y0, #16
		sar	y0, #16
		mov	x0, x0y0
		sar	x0, #16


		test	rad, H8000  wz	' test if fill or draw
		and	rad, LOW_UWORD
	if_z	jmp	#draw_circle

fill_circle
'		and	x0, H3FF
'		and	y0, H3FF
'		and 	rad, H3FF
		
		mov	rr, rad		' drr is negative sense, = rad^2 - (x^2 + y^2) + offset
		mov	dxx, #1
		mov	dyy, rad
		add	dyy, rad
		add	dyy, #1
		mov	x1, #0
		mov	y1, rad
:loop
		cmpsub	rr, dxx  wc	' is this right sense, skip if rr >= dxx
	if_nc	jmp	#:skip
		add	x1, #1
		add	dxx, #2
		jmp	#:loop
:skip
		mov	span_len, x1
		add	span_len, x1
		add	span_len, #1

		neg	xadj, x1
		neg	yadj, y1
		call	#queue_hspan	'hspan   x0-x1, y0-y1, x1+x1+1, 

		mov	yadj, y1
		call	#queue_hspan	'hspan   x0-x1, y0+y1, x1+x1+1, 

		sub     dyy, #2
		add	rr, dyy
		djnz	y1, #:loop

			
		mov	span_len, rad
		add	span_len, rad
		add	span_len, #1

		neg	xadj, rad
		mov	yadj, #0
		call	#queue_hspan 	'hspan   x0-rad, y0, rad+rad+1, 

		jmp	#circle_done

draw_circle

		mov	rr, rad		' drr is negative sense, = rad^2 - (x^2 + y^2) + offset
		mov	dxx, #1
		mov	dyy, rad
		add	dyy, rad
		add	dyy, #1
		mov	x1, #0
		mov	y1, rad

		sub	x0, rad
		call	#draw_dot
		add	x0, rad
		add	x0, rad
		call	#draw_dot
		sub	x0, rad
		sub	y0, rad
		call	#draw_dot
		add	y0, rad
		add	y0, rad
		call	#draw_dot
		sub	y0, rad

:yloop		
		mov	xp, x1
:loop
		cmpsub	rr, dxx  wc	' is this right sense, skip if rr >= dxx
	if_nc	jmp	#:skip
		add	x1, #1
		add	dxx, #2
		jmp	#:loop
:skip
		sub	xp, x1
		neg	xp, xp  wz
	if_z	jmp	#:skipspan      ' probably cant happen
		
		mov	span_len, xp

		neg	xadj, x1
		neg	yadj, y1
		call	#queue_hspan
		mov	yadj, y1
		call	#queue_hspan

		mov	xadj, x1
		sub	xadj, xp
		add	xadj, #1
		call	#queue_hspan
		neg	yadj, y1
		call	#queue_hspan


		neg	xadj, y1
		neg	yadj, x1
		call	#queue_vspan
		mov	xadj, y1
		call	#queue_vspan

		mov	yadj, x1
		sub	yadj, xp
		add	yadj, #1
		call	#queue_vspan
		neg	xadj, y1
		call	#queue_vspan

:skipspan
		sub     dyy, #2
		add	rr, dyy
		sub	y1, #1

		cmps	y1, x1  wz,wc
	if_ae	jmp	#:yloop


circle_done
		sub	outqp, outq_beg
		shr	outqp, #2
		wrlong  outqp, outqp_addr   ' write back queue insert pointer cache before we mark circle as finished
mark_done
		mov	rad, #0
		wrlong	rad, parm           ' mark as finished
		jmp	#workloop

{ ---------------------------------------------------------- }

draw_dot
{
		cmps	x0, clip_left  wz,wc ' clip
	if_ae	cmps	clip_right, x0  wz,wc
	if_be	jmp	#draw_dot_ret
		cmps	y0, clip_top  wz,wc
	if_ae	cmps	clip_bot, y0  wz,wc
	if_be	jmp	#draw_dot_ret
}
		mov	xxyy, x0
		shl	xxyy, #16
		mov	tmp, y0
		and	tmp, LOW_WORD
		or	xxyy, tmp
		and	xxyy, H1FFFFFFF
		or	xxyy, tag_vspan
		and	spanetc, LOW_WORD
		or	spanetc, H10000
		call	#fastenq
draw_dot_ret	ret

H10000		long	$10000
{ ---------------------------------------------------------- }

'clip_top	long    100
'clip_bot	long	HEIGHT-50
'clip_left	long	50
'clip_right	long	WIDTH-250



queue_hspan
		mov	tmp, y0			   ' y compute
		add	tmp, yadj
{
		cmps	tmp, clip_top  wz,wc       ' y clip test
	if_ae	cmps	clip_bot, tmp  wz,wc
	if_be	jmp	#queue_hspan_ret
}
	        mov	xxyy, x0		   ' x compute
		add	xxyy, xadj

		mov	clipped_span, span_len      ' x span-clipping
{
		add	clipped_span, xxyy	    ' compute span right
		mins	xxyy, clip_left
		maxs	clipped_span, clip_right
		cmps	clipped_span, xxyy  wz,wc,wr
	if_be	jmp	#queue_hspan_ret 
}
		shl	xxyy, #16		    ' generate the two encoded longs
		and	tmp, LOW_WORD
		or	xxyy, tmp

		and	xxyy, H1FFFFFFF
		or	xxyy, tag_hspan

		and	spanetc, LOW_WORD
		shl	clipped_span, #16
		or	spanetc, clipped_span
		call	#fastenq
queue_hspan_ret	ret

queue_vspan     mov	xxyy, x0		    ' x compute
		add	xxyy, xadj
{
		cmps	xxyy, clip_left  wz,wc	    ' x clip test
	if_ae	cmps	clip_right, xxyy  wz,wc
	if_be	jmp	#queue_vspan_ret
}
		mov	tmp, y0			    ' y compute
		add	tmp, yadj

		mov	clipped_span, span_len      ' y span-clip
{
		add	clipped_span, tmp	    ' compute span bot
		mins	tmp, clip_top
		maxs	clipped_span, clip_bot
		cmps    clipped_span, tmp  wz,wc,wr
	if_be	jmp	#queue_vspan_ret
}
		shl	xxyy, #16		   ' generate two encoded longs
		and	tmp, LOW_WORD
		or	xxyy, tmp

		and	xxyy, H1FFFFFFF
		or	xxyy, tag_vspan

		and	spanetc, LOW_WORD
		shl	clipped_span, #16
		or	spanetc, clipped_span
		call	#fastenq
queue_vspan_ret	ret

{ ---------------------------------------------------------- }

{ ---------------------------------------------------------- }



' enqueue an entry with xxyy, spanetc
fastenq		mov	outqp2, outqp
		rdlong	tmp, outqp  wz	' examine entry at insert point
	if_nz	jmp	#fastenq
		wrlong	xxyy, outqp     ' fill up the entry
		add	outqp2, #4
		add	outqp, #8
		wrlong  spanetc, outqp2
		cmp	outqp, outq_end  wz,wc
	if_ae	mov	outqp, outq_beg
fastenq_ret	ret

{ ---------------------------------------------------------- }

H3FF		long    $3FF
LOW_WORD	long	$0000FFFF
LOW_UWORD	long	$00007FFF
H8000		long	$8000
H1FFFFFFF	long	$1FFFFFFF
queue_bytes	long	Q_BYTES
tag_hspan	long    OP_HSPAN << 29
tag_vspan	long    OP_VSPAN << 29

parm		res	1

outq_beg	res	1
outq_end	res	1
outqp		res	1
outqp2		res	1
outqp_addr	res	1
tmp		res	1
xxyy		res	1
spanetc		res	1

rad		res	1
xp		res	1
x0		res	1
y0		res	1
x0y0		res	1
x1		res	1
y1		res	1
xadj		res	1
yadj		res	1
span_len	res	1
clipped_span	res	1
rr		res	1
dxx		res	1
dyy		res	1

		FIT	$1F0

{{
////////////////////////////////////////////////////////////////////////////////////////////
//                                TERMS OF USE: MIT License
////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this 
// software and associated documentation files (the "Software"), to deal in the Software 
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////////////////
}}
