{{ sdram-video-raster.spin }}
CON
  Q_SIZE = 400
  Q_BYTES = Q_SIZE << 3   ' entries are 2 longs

  WIDTH = 800
  HEIGHT = 480


  OP_NOP   = 0
  OP_HSPAN = 1
  OP_VSPAN = 2
  OP_CHAR  = 3
  OP_SPRITE = 4

  ' polygon stuff

  y1_off = 0  ' unsigned
  y0_off = 2  ' signed
  dx_off = 4  ' unsigned
  dy_off = 6  ' signed
  xx_off = 8  ' signed
  h_off	 = 10 ' signed

  ' the vertex table is x, y (lsw, msw)

  EDGE_SIZE = 6

  MAX_EDGES = 160

VAR
  word edge_table [MAX_EDGES * EDGE_SIZE]  ' about 2k RAM

  long polygon_cog


PUB  Start (args)
  Stop
  long [args][3] := 0  ' vcount
  long [args][6] := @edge_table
  polygon_cog := 1 + cognew (@polygon, args) 'Launch new cog
  if polygon_cog == 0
    Stop
    return false
  else
    return true

PUB  Stop
  if polygon_cog <> 0
    cogstop (polygon_cog-1)
    polygon_cog := 0


DAT
		ORG	0

polygon		mov	parm, PAR
		rdlong	outq_beg, parm
		add	parm, #8
		mov	outqp_addr, parm
		add	parm, #16
		rdlong	edge_table_addr, parm
		sub	parm, #12

		mov	outq_end, outq_beg
		add	outq_end, queue_bytes

workloop	rdlong	vcount, parm  wz   ' our workqueue, first a vertex count
	if_z	jmp	#workloop
		rol	vcount, #1  wc
		shr	vcount, #1
		muxc	mode, #1
		cmp	vcount, #3  wz,wc
	if_b	jmp	#:skipwork	    ' less than 3 edges is invisible

		rdlong	outqp, outqp_addr   ' get the queue insert pointer
		add	parm, #4
		shl	outqp, #2
		rdlong	spanetc, parm      ' install colour into spanetc
		and	spanetc, LOW_PART
		add	parm, #4
		rdlong	vert_table_addr, parm       ' get vertex table
		sub	parm, #8	    ' back to point at arg_vcount
		add	outqp, outq_beg
		
		call	#poly_rend

		sub	outqp, outq_beg
		shr	outqp, #2
		wrlong	outqp, outqp_addr   ' return the queue insert pointer
:skipwork
		wrlong  zer, parm       ' flag we're done enqueuing results and ready for more
		jmp	#workloop


queue_bytes	long	Q_BYTES


{ ------------------------------------------- }

poly_rend
		mov	ep, #edge_start
		mov	etabp, edge_table_addr
		mov	counter, vcount
		shl	vcount, #2	' bytesize of vert table, 4 bytes per entry
		mov	ecount, #0
		mov	maxy, #1
		shl	maxy, #31	' maxy = 80000000
		mov	miny, maxy
		sub	miny, #1	' miny = 7FFFFFFF

{ vertex loop scans the vertex array and adds all the edges to the edge table }
vert_loop
		rdlong	x0, vert_table_addr
		add	vert_table_addr, #4
		mov	y0, x0
		shl	x0, #16		     ' x is signed lsw
		sar	x0, #16
		sar	y0, #16	             ' y is signed msw
		cmp	counter, #1  wz	     ' test for last entry
	if_z	sub	vert_table_addr, vcount  ' if so wrap round for following vertex
		rdlong	x1, vert_table_addr
		mov	y1, x1
		shl	x1, #16
		sar	x1, #16
		sar	y1, #16

		cmps	y0, y1  wc,wz
	if_z	jmp	#:skip_edge	     ' don't do anything for horizontal edges
{ reverse if necessary }
	if_b	jmp	#:skip_rev
		mov	t, x0		     ' reverse line so y1 > y0
		mov	x0, x1
		mov	x1, t
		mov	t, y0
		mov	y0, y1
		mov	y1, t
:skip_rev

{ y clipping }
		cmps	y1, clip_top  wz,wc ' entirely above?
	if_b	jmp	#:skip_edge
		cmps	y0, clip_bot  wz,wc     ' entirely below?
	if_ae	jmp	#:skip_edge

{ track limits }
		maxs	miny, y0	     ' compute limits
		mins	maxy, y1
		
{ write out the edge }

  	    	cmp	ecount, #MAX_EDGES  wz,wc ' perhaps edge table full?
	if_ae	jmp	#return_point

		movd	:ins1, ep       ' add to register array
		add	ep, #1 		' move pointer
		add	ecount, #1	' and count, recording edge
:ins1		mov	0-0, etabp
		wrword	y1, etabp	' unsigned after clipping
		add	etabp, #2
		wrword	y0, etabp
		add	etabp, #2
		sub	y1, y0		' dx = y1-y0
		wrword	y1, etabp	' unsigned after reversing
		add	etabp, #2
		subs	x1, x0  wz,wc	' dy = x1-x0
		wrword	x1, etabp
		add	etabp, #2
		wrword	x0, etabp	' xx signed
		add	etabp, #2
		abs	x1, x1		' h is derived from abs(dy)
		shr	x1, #1
		neg	x1, x1
		wrword	x1, etabp	' h is signed
		add	etabp, #2

:skip_edge
		djnz	counter, #vert_loop

		mov	edge_last, ep	' record last edge pointer

{ now have all the edges in table, time to see if any and sort them in y0 }

		cmp	ecount, #0  wz 	' if no edges in table then all were off screen or horizontal
	if_z	jmp	#return_point

		maxs	maxy, clip_bot	' ensure we stop at bot of screen
		mov	yy, miny	' have to start above clip-rect if necessary

		mov	field, #y0_off  ' sort by y0 
		mov	frm, #edge_start
		mov	too, edge_last
		call	#sort

		mov	tt, #edge_start
		mov	ff, #edge_start


{ render loop: maintain active set, progress edges, sort on xx, do spans }
render_loop
		cmps	yy, maxy  wz,wc
 	if_ae	jmp	#return_point




{ tt loop scans for new edges to make active, already have end of array sorted in y0 order }

:tt_loop	cmp	tt, edge_last  wz,wc  ' have to avoid going off end of array
		movs	:ldins, tt
	if_ae	jmp	#:skip_tt_loop

:ldins		rdlong	y0, 0-0	        ' y0 is msw in long 0
		sar	y0, #16

		cmp	y0, yy  wz
	if_z	add	tt, #1		' y0 == yy means add to active
	if_z	jmp	#:tt_loop
:skip_tt_loop


{ ff loop scans active edges for any that have reached conculsion (y1 == yy) }

		mov	ii, ff
ff_loop	        cmp	ii, tt  wc,wz
	  	movs	:ldins, ii
	if_ae	jmp	#end_ff_loop

:ldins		rdword	t, 0-0		' y1 is word 0, unsigned

		cmp	t, yy  wz
	if_nz	jmp	#:skip_ff

		cmp	ii, ff  wz	' want to lose entry at i, copy ff to i
	if_z	jmp	#:skip_ff2
		movs	:ins, ff
		movd	:ins, ii
		nop
:ins		mov	0-0, 0-0
:skip_ff2
		add	ff, #1		' move ff on
:skip_ff
		add	ii, #1
		jmp	#ff_loop
end_ff_loop

{ advancing edges }

		mov	ii, ff
:rloop		cmp	ii, tt  wc,wz
	if_ae	jmp	#:end_rloop
		call	#advance_edge
		add	ii, #1
		jmp	#:rloop
:end_rloop

{ now the check for above screen still } 

		cmps	yy, clip_top  wz,wc  ' ignore if above screen
	if_b	jmp	#:skip_spans

{ x sort }
		mov	field, #xx_off ' sort x's
		mov	frm, ff
		mov	too, tt
		call	#sort

{ do spans }

		mov	ii, ff
:span_loop	cmp	ii, tt  wz,wc
	if_ae	jmp	#:end_span_loop
		
		call	#do_span

		add	ii, #2   ' spans are between a pair of edges, using even-odd winding rule
		jmp	#:span_loop
:end_span_loop

:skip_spans
		add	yy, #1
		jmp	#render_loop

return_point
poly_rend_ret	ret

{ ---------------------------------------------------------- }

' takes field, frm, too, corrupts i, j, t, change
sort		mov	ii, too
		sub	ii, #1
:sloop1
		mov	jj, frm
		mov	change, #0
:sloop2
		call	#compare
	if_be	add	jj, #1
	if_be	jmp	#:skipswap
                movs	:ins1, jj
		movd	:ins2, jj
		add	jj, #1		' have to inc j anyway
		movs	:ins2, jj
		movd	:ins3, jj
:ins1		mov	t, 0-0
:ins2		mov	0-0, 0-0
:ins3		mov	0-0, t
		mov	change, #1

:skipswap
		cmp	jj, ii  wz
	if_nz	jmp	#:sloop2

		test	change, #1  wz
	if_z	jmp	#sort_ret

		sub	ii, #1
		cmp	ii, frm  wz
	if_nz	jmp	#:sloop1
sort_ret	ret

{ ---------------------------------------------------------- }

' takes i as the edge to process
advance_edge	movs	:ins, ii
		nop
:ins		mov	ep, 0-0
		'rdlong	y1, ep		' not needing y0 or y1 here
		'mov	y0, y1
		add	ep, #4
		rdlong	dx, ep		' load dx dy
		add	ep, #4
		mov	dy, dx
		rdlong	xx, ep		' load xx h
		mov	hh, xx
		sar	dy, #16		' dy is signed
		and	dx, LOW_PART	' dx isn't
		sar	hh, #16		' h is signed
		shl	xx, #16		' xx is signed
		sar	xx, #16

		addabs	hh, dy		' h += |dy|
:xloop
		cmps	hh, #0  wc,wz    ' while (h > 0)
	if_be	jmp	#:wrback
		sub	hh, dx		' { h -= dx
		cmps	dy, #0  wc,wz
	if_b	sub	xx, #1
	if_ae	add	xx, #1          '   xx += sign (dy)
		jmp	#:xloop         ' }

:wrback		shl	hh, #16
		and	xx, LOW_PART
		or	hh, xx
		wrlong	hh, ep
advance_edge_ret ret

{ ---------------------------------------------------------- }
{{
' takes from and to, overwrites t
swap	        movs	:ins1, frm
		movd	:ins2, frm
		movs	:ins2, too
		movd	:ins3, too
:ins1		mov     t, 0-0
:ins2		mov	0-0, 0-0
:ins3		mov	0-0, t
swap_ret	ret
}}
{ ---------------------------------------------------------- }


' takes field as the offset, j and j+1 as the edges.
compare	        movs    :ins1, jj
		movs	:ins2, jj
		add	:ins2, #1
:ins1		mov	t, 0-0 
		add	t, field
		rdword	t, t
:ins2		mov	t2, 0-0
		add	t2, field
		rdword	t2, t2
		shl	t, #16
		shl	t2, #16		' make signed, can leave shifted left for comparison
		cmps	t, t2  wz, wc
compare_ret     ret

{ ---------------------------------------------------------- }


' takes i and i+1 as edges to generate span
do_span		movs	:ins1, ii
		movs	:ins2, ii
		add	:ins2, #1
:ins1		mov	t, 0-0
		add	t, #xx_off
		rdword	t, t
:ins2		mov	t2, 0-0
		add	t2, #xx_off
		rdword	t2, t2
		shl	t, #16
		shl	t2, #16
		test	mode, #1  wc
	if_c	jmp	#do_dots
		sar	t, #16		' xx's are signed
		sar	t2, #16
		cmps	t2, t  wz,wc,wr	' calculate span and compare to empty
		' here t = span start, t2 = span width, yy = y
	if_a	call    #rend_span     ' probably claims a lock and enqueues, or else our caller waits.
do_span_ret	ret

do_dots		mov	xxyy, t
		or	xxyy, yy
		and	xxyy, H1FFFFFFF
		or	xxyy, tag_vspan
		and	spanetc, LOW_PART
		or	spanetc, H10000
		call	#fastenq
		sub	xxyy, t
		add	xxyy, t2
		and	xxyy, H1FFFFFFF
		or	xxyy, tag_vspan
		call	#fastenq
		jmp	#do_span_ret

{ ---------------------------------------------------------- }

rend_span	mov	xxyy, t
		shl	xxyy, #16
		or	xxyy, yy
		and	xxyy, H1FFFFFFF
		or	xxyy, tag_hspan
		and	spanetc, LOW_PART ' just colour
		shl	t2, #16
		or	spanetc, t2
		call	#fastenq
rend_span_ret	ret

H1FFFFFFF	long	$1FFFFFFF

{ ---------------------------------------------------------- }


' enqueue an entry with xxyy, spanetc
fastenq		mov	outqp2, outqp
		rdlong	tmp, outqp  wz	' examine entry at insert point
	if_nz	jmp	#fastenq

		wrlong	xxyy, outqp     ' fill up the entry
		add	outqp2, #4
		add	outqp, #8
		wrlong  spanetc, outqp2
		cmp	outqp, outq_end  wz,wc
	if_ae	mov	outqp, outq_beg
fastenq_ret	ret

{ ---------------------------------------------------------- }


zer		long	0

clip_top	long	0
clip_bot	long	$1000 ' HEIGHT
'clip_left	long	300
'clip_right	long	WIDTH-50

tag_hspan	long	OP_HSPAN << 29
tag_vspan	long	OP_VSPAN << 29

LOW_PART        long    $0000FFFF
H8000           long    $8000
H10000		long	$10000

parm            res     1


x0		res	1
y0		res	1
x1		res	1
y1		res	1
miny		res	1
maxy		res	1
inc             res     1
yy              res     1
dx              res     1
dy              res     1
xx              res     1
hh              res     1

ii		res	1
jj              res     1
vcount		res	1
ecount		res	1
t		res	1
t2              res     1
etabp		res	1
ep		res	1		
ff		res	1
tt		res	1
frm             res     1
too             res     1
edges_last	res	1
vert_table_addr	res	1
edge_table_addr	res	1
counter		res	1
edge_last       res     1
field           res     1
change          res     1
mode		res	1

		' table of pointers to edge structs, we just re-order this table when sorting
edge_start	res	MAX_EDGES
edge_end




outq_beg	res	1
outq_end	res	1
outqp		res	1
outqp2		res	1
outqp_addr	res	1
tmp		res	1
'vtable		res	1
xxyy		res	1
spanetc		res	1
		

		FIT	$1F0

{{
////////////////////////////////////////////////////////////////////////////////////////////
//                                TERMS OF USE: MIT License
////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this 
// software and associated documentation files (the "Software"), to deal in the Software 
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////////////////
}}
